import requests
import uuid
from argparse import ArgumentParser
from babbagecoin.common.models import Transaction, SignedTransaction
from babbagecoin.common.wallet import Wallet
from babbagecoin.common.schemas import SignedTransactionSchema
from babbagecoin.common.context import NetworkContext


parser = ArgumentParser(description="Send a transaction on babbagecoin")

parser.add_argument("receiver", type=str)
parser.add_argument("amount", type=float)
parser.add_argument("fees", type=float)

context = NetworkContext()


class Client:
    wallet: Wallet

    def __init__(self, load_from_file=True):
        self.wallet = Wallet()

    def send_transaction(self, receiver: str, amount: float, fees: float):
        print(f"Sending {amount} to {receiver} with fees {fees}")
        tx = Transaction(
            uuid=str(uuid.uuid4()),
            sender=self.wallet.get_public_key(),
            receiver=receiver,
            amount=float(amount),
            fees=float(fees),
        )

        signedTx = SignedTransaction(tx, self.wallet.sign(tx))
        try:
            requests.post(
                f"{context.myUrl}/transactions/emit",
                SignedTransactionSchema.dumps(signedTx),
                headers={"Content-Type": "application/json"},
            )
        except requests.exceptions.ConnectionError:
            print(f"Node master at {context.myUrl} is not accessible.")

    def get_balance(self):
        try:
            addr = str(self.wallet.get_public_key())
            res = requests.get(
                f"{context.myUrl}/addresses/{addr}/balance",
                headers={"Content-Type": "application/json"},
            )
            print(res.json())
        except requests.exceptions.ConnectionError:
            print(f"Node master at {context.myUrl} is not accessible.")


def check_balance():
    client = Client()
    client.get_balance()


def send_transaction(receiver, amount, fees: float):
    client = Client()
    client.send_transaction(receiver, amount, fees)



import os
from dotenv import dotenv_values


def get_env_variables():
    return dotenv_values(".env")


def get_current_user():
    if "CURRENT_USER" in get_env_variables():
        return get_env_variables()["CURRENT_USER"]
    else:
        return "NO_CONFIG_USER"


def get_my_ip():
    return get_env_variables()[f"IP_NODE_{get_current_user()}"]


def get_sentry_dsn():
    """
    Return the sentry DSN for monitoring of logs using glitchtip (open source clone of sentry)
    """
    envs = get_env_variables()
    if "DSN_GLITCHTIP" in envs:
        return get_env_variables()["DSN_GLITCHTIP"]
    return None


def isIP(env_var):
    name, value = env_var
    if "IP_NODE" in name and value:
        return True
    return False


def get_all_ips():
    ips = []
    for _, ip in filter(isIP, get_env_variables().items()):
        ips.append(ip)
    return ips


class NetworkContext:
    """Implement a singleton. NetworkContext() will always returns the exact
    same object instance."""

    instance: "NetworkContext" = None

    def __init__(self):
        if NetworkContext.instance is None:
            if os.environ.get("TESTING") or not os.path.isfile(".env"):
                self.myIp = "127.0.0.1"
                self.known_nodes = [self.myIp]
                self.sentry_dsn = None
            else:
                self.myIp = get_my_ip()
                self.known_nodes = get_all_ips()
                self.sentry_dsn = get_sentry_dsn()

            self.myUrl = f"http://{self.myIp}:5000"

            NetworkContext.instance = self

    def __new__(cls):
        if cls.instance:
            return cls.instance
        return super().__new__(cls)


if __name__ == "__main__":
    print(id(NetworkContext().instance))
    print(id(NetworkContext()))
    a = NetworkContext()
    print(id(a))
    # always the same object



import uuid

from babbagecoin.common.models import (
    MINING_REWARD_AMOUNT,
    MINING_REWARD_ADDRESS,
    PubKey,
    Block,
    SignedTransaction,
    Transaction,
)
from babbagecoin.common.wallet import wallet
from babbagecoin.common.balance import apply_transaction, cancel_transaction

reward_transaction = None
mem_pool = set()
validated_transactions = set()
reward_transaction: Transaction = None


def update_block_transactions(block: Block):
    global mem_pool
    block.signed_transactions.extend(list(mem_pool))


def refresh_transactions_from_new_block(block: Block):
    global mem_pool
    global validated_transactions
    excess_transactions = set()
    miner = block.signed_transactions[0].transaction.receiver
    for stx in block.signed_transactions:
        if stx in validated_transactions:
            excess_transactions.add(stx)
        else:
            validated_transactions.add(stx)
            apply_transaction(miner, stx)
        if stx in mem_pool:
            mem_pool.remove(stx)
    return excess_transactions


def refresh_transactions_from_old_block(block: Block):
    global mem_pool
    global validated_transactions
    miner = block.signed_transactions[0].transaction.receiver
    for stx in block.signed_transactions:
        if stx in validated_transactions:
            validated_transactions.remove(stx)
        cancel_transaction(miner, stx)
        if stx.transaction.sender.dumps() != MINING_REWARD_ADDRESS:
            mem_pool.add(stx)


def forge_reward_transaction() -> SignedTransaction:
    transaction = Transaction(
        uuid=str(uuid.uuid4()),
        sender=PubKey(MINING_REWARD_ADDRESS),
        receiver=wallet.get_public_key().hash(),
        amount=MINING_REWARD_AMOUNT,
    )
    return SignedTransaction(
        transaction=transaction,
        signature=wallet.sign(transaction),
    )


reward_transaction = forge_reward_transaction()


def update_reward_transaction():
    global reward_transaction
    reward_transaction = forge_reward_transaction()


def get_reward_transaction():
    return reward_transaction



import json
import requests
import sentry_sdk
from sentry_sdk.integrations.flask import FlaskIntegration


from flask import Flask, request

from babbagecoin.common.schemas import BlockSchema, SignedTransactionSchema
from babbagecoin.master.blockchain_service import (
    get_head,
    block_tbl,
    build_working_block,
    make_primary_between,
    update_blockchain,
)
from babbagecoin.common.context import NetworkContext
from babbagecoin.common.balance import get_balance_of_address
from babbagecoin.master.broadcast_service import broadcast_block, broadcast_transaction
from babbagecoin.master.transaction_service import (
    update_block_transactions,
    update_reward_transaction,
    mem_pool,
)


sentry_sdk.init(dsn=NetworkContext().sentry_dsn, integrations=[FlaskIntegration()])

app = Flask(__name__, static_folder="../webclient/")


@app.get("/")
def print_chain():
    s = ""
    b = get_head()
    while b.height != 0:
        s += b.html()
        b = block_tbl[b.prev_hash]
    return s


@app.get("/greetings")
def hello():
    return {"message": "greetings"}


@app.get("/webclient")
def get_client_html():
    return app.send_static_file("index.html")


@app.get("/blocks/working_block")
def get_working_block():
    next_block = build_working_block()
    update_block_transactions(next_block)
    json_block = BlockSchema.dumps(next_block)
    return json_block


@app.get("/blocks/<block_hash>")
def get_block_with_hash(block_hash: str):
    json_block = BlockSchema.dumps(block_tbl[block_hash])
    return json_block


@app.get("/blocks/hashdict")
def send_block_tbl():
    return block_tbl


@app.post("/blocks/minedblock")
def receive_mined_block_from_miner():
    mined_block = BlockSchema.load(request.json)
    update_blockchain(mined_block, mined_block)
    broadcast_block(mined_block)
    update_reward_transaction()
    return "ok"


@app.post("/blocks/updateblock")
def receive_block_from_network():
    leaf = BlockSchema.load(request.json["block"])
    url = request.json["url"]
    b = leaf
    if leaf.hash() not in block_tbl:
        while b.prev_hash not in block_tbl:
            print(f"Fetching block: {b.prev_hash}")
            resp = requests.get(
                f"{url}/blocks/{b.prev_hash}",
                headers={"Content-Type": "application/json"},
            )
            prev_b = BlockSchema.load(resp.json())
            make_primary_between(prev_b, b)
            b = prev_b
        update_blockchain(b, leaf)
    return "ok"


@app.post("/transactions/emit")
def emit_transaction():
    signed_transaction = SignedTransactionSchema.load(request.json)
    mem_pool.add(signed_transaction)
    broadcast_transaction(signed_transaction)
    return "ok"


@app.post("/transactions/add")
def add_transaction():
    signed_transaction = SignedTransactionSchema.load(request.json)
    mem_pool.add(signed_transaction)
    return "ok"


@app.get("/addresses/<address>/balance")
def get_balance(address: str):
    balance = get_balance_of_address(address)
    return json.dumps({"balance": balance, "address": address})



from babbagecoin.common.models import MINING_REWARD_ADDRESS
from babbagecoin.common.models import SignedTransaction, SUCCESS, REVERTED

balances = {}


def touch_balance(addr: str):
    global balances
    if addr not in balances:
        balances[addr] = 0


def get_balance_of_address(addr: str):
    touch_balance(addr)
    return balances[addr]


def apply_transaction(miner: str, stx: SignedTransaction):
    global balances
    tx = stx.transaction
    sender = tx.sender.hash()
    receiver = tx.receiver
    touch_balance(sender)
    touch_balance(receiver)
    touch_balance(miner)
    new_sender_balance = balances[sender] - tx.amount - tx.fees
    if new_sender_balance >= 0 or tx.sender.dumps() == MINING_REWARD_ADDRESS:
        balances[sender] = new_sender_balance
        balances[receiver] += tx.amount
        balances[miner] += tx.fees
        tx.status = SUCCESS
    else:
        tx.status = REVERTED


def cancel_transaction(miner: str, stx: SignedTransaction):
    global balances
    tx = stx.transaction
    sender = tx.sender.hash()
    receiver = tx.receiver
    if tx.status == SUCCESS:
        balances[sender] += tx.amount + tx.fees
        balances[receiver] -= tx.amount
        balances[miner] -= tx.fees
    else:
        assert tx.status == REVERTED



from typing import Dict, Any

from marshmallow import Schema, fields, post_load

from babbagecoin.common.models import PubKey, Transaction, SignedTransaction, Block


class PubKeyField(fields.Field):
    def _serialize(self, value: PubKey, attr: str, obj: Any, **kwargs) -> str:
        return bytes.hex(value.dump())

    def _deserialize(self, value: str, attr, data, **kwargs) -> PubKey:
        return PubKey.load_from_bytes(bytes.fromhex(value))


class _TransactionSchema(Schema):
    uuid = fields.String()
    sender = PubKeyField()
    receiver = fields.String()
    amount = fields.Float()
    fees = fields.Float()

    @post_load
    def _make_model(self, data: Dict[str, Any], **kwargs) -> Transaction:
        return Transaction(**data)


TransactionSchema = _TransactionSchema()


class _SignedTransactionSchema(Schema):
    transaction = fields.Nested(TransactionSchema)
    signature = fields.String()

    @post_load
    def _make_model(self, data: Dict[str, Any], **kwargs) -> SignedTransaction:
        return SignedTransaction(**data)


SignedTransactionSchema = _SignedTransactionSchema()


class _BlockSchema(Schema):
    prev_hash = fields.String()
    height = fields.Int()
    signed_transactions = fields.List(fields.Nested(SignedTransactionSchema))
    nonce = fields.Int(required=False)

    @post_load
    def _make_model(self, data: Dict[str, Any], **kwargs) -> Block:
        return Block(**data)


BlockSchema = _BlockSchema()



from babbagecoin.common.models import Block


difficulty: int = None
bound: int = None


def set_difficulty(new_difficulty: int):
    global difficulty, bound
    difficulty = new_difficulty
    bound = 2**256 // difficulty


set_difficulty(2000000)


def is_block_hash_valid(block: Block):
    return int(block.hash(recompute=True), 16) <= bound



import time
import random
import requests

from typing import Optional

from babbagecoin.common.models import Block
from babbagecoin.common.schemas import BlockSchema
from babbagecoin.common.block_service import is_block_hash_valid
from babbagecoin.common.context import NetworkContext

context = NetworkContext()


class BlockStore:
    working_block: Optional[Block] = None
    initial_hash: Optional[str] = None

    def update_working_block(self):
        new_block = get_working_block()
        new_initial_hash = new_block.hash()

        if self.initial_hash != new_initial_hash:
            print("Updating working block")
            new_block.nonce = random.randint(0, 2**64)
            self.working_block = new_block
            self.initial_hash = new_initial_hash
        else:
            print("Keeping same working block")


def get_working_block() -> Block:
    res = requests.get(f"{context.myUrl}/blocks/working_block")
    # print(res.json())
    return BlockSchema.load(res.json())


def post_mined_block(block: Block):
    json_block = BlockSchema.dumps(block)

    try:
        requests.post(
            f"{context.myUrl}/blocks/minedblock",
            json_block,
            headers={"Content-Type": "application/json"},
        )
    except Exception as e:
        print(f"Could not send mined block: {e}")


def run():
    time.sleep(0.5)  # waiting for the master to be up
    block_store = BlockStore()
    block_store.update_working_block()

    hash_count_before_update = 1000000
    i = 0
    while i < hash_count_before_update:
        if is_block_hash_valid(block_store.working_block):
            print("BLOCK MINED SUCCESSFULLY")
            post_mined_block(block_store.working_block)
            block_store.update_working_block()
            i = 0
        else:
            block_store.working_block.nonce += 1
            i += 1
        if i == hash_count_before_update:
            block_store.update_working_block()
            i = 0

    print("======= WORKER TERMINATED EARLY ========")



class RewardTransactionNotUnique(Exception):
    pass


class InvalidBlockHash(Exception):
    """The block hash does not respect the difficulty constraint"""

    pass


class InvalidBlockHeight(Exception):
    pass


class InvalidSignatureForTransaction(Exception):
    pass


class BadRewardTransaction(Exception):
    pass



import sys
from babbagecoin.miner.miner import run as miner_run
from babbagecoin.master.routes import app as master_app
from babbagecoin.client.client import send_transaction, check_balance


if __name__ == "__main__":
    command = sys.argv[1]
    if command == "master":
        master_app.run(debug=True, host="0.0.0.0")
    elif command == "miner":
        miner_run()
    elif command == "tx":
        send_transaction(*sys.argv[2:])
    elif command == "balance":
        check_balance()



import hashlib

from dataclasses import dataclass, field
from typing import List, Optional
from cryptography.hazmat.backends.openssl.rsa import RSAPublicKey
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import load_pem_public_key

MINING_REWARD_AMOUNT = 100
MINING_REWARD_ADDRESS = "BABBAGE"

SUCCESS = 0
REVERTED = 1


@dataclass
class PubKey:
    rsa_pub_key: RSAPublicKey or str

    def dump(self) -> bytes:
        if self.rsa_pub_key == MINING_REWARD_ADDRESS:
            return MINING_REWARD_ADDRESS.encode()
        return self.rsa_pub_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )

    def dumps(self):
        return self.dump().decode()

    def hash(self) -> str:
        if self.rsa_pub_key == MINING_REWARD_ADDRESS:
            return MINING_REWARD_ADDRESS
        encoded_key = self.dump()
        hasher = hashlib.sha256()
        hasher.update(encoded_key)
        return hasher.hexdigest()

    def __str__(self):
        return self.hash()

    @staticmethod
    def load_from_bytes(rsa_pk: bytes):
        if rsa_pk.decode() == MINING_REWARD_ADDRESS:
            return PubKey(MINING_REWARD_ADDRESS)
        return PubKey(load_pem_public_key(rsa_pk))


@dataclass
class Transaction:
    uuid: str
    sender: PubKey
    receiver: str
    amount: float
    fees: float = 0
    status: int = SUCCESS  # REVERTED when insufficient funds

    def hash(self):
        hasher = hashlib.sha256()
        hasher.update(self.uuid.encode())
        hasher.update(self.sender.dump())
        hasher.update(self.receiver.encode())
        hasher.update(str(self.amount).encode())
        hasher.update(str(self.fees).encode())
        return hasher.hexdigest()

    def html(self):
        uuid = self.uuid[:4]
        sender = str(self.sender)[:8]
        receiver = str(self.receiver)[:8]
        amount = str(self.amount)
        fees = str(self.fees)
        status_str = ""
        if self.status == REVERTED:
            status_str = "<b>REVERTED</b> "
        return (
            f"{status_str}"
            f"uuid: {uuid} "
            f"sender: {sender} "
            f"receiver: {receiver} "
            f"amount: {amount} "
            f"fees: {fees}"
        )

    def __repr__(self) -> str:
        return f"Transaction({self.html()})"


@dataclass
class SignedTransaction:
    transaction: Transaction
    signature: str

    def hash(self):
        hasher = hashlib.sha256()
        hasher.update(self.transaction.hash().encode())
        hasher.update(self.signature.encode())
        return hasher.hexdigest()

    # __hash__() and __eq__() should be defined for
    # sets of SignedTransaction to work properly
    def __hash__(self):
        return int(self.hash(), 16)

    def __eq__(self, stx: "SignedTransaction"):
        return self.hash() == stx.hash()

    def __str__(self):
        return f"stx({self.transaction})"

    def html(self):
        return self.transaction.html()


@dataclass
class Block:
    height: int
    prev_hash: Optional[str] = None
    # Invariant: the first transaction should be the reward transaction for mining the block
    signed_transactions: List[SignedTransaction] = field(default_factory=list)
    nonce: int = 0

    # The following data is not hashed
    # Head path invariant: following the first element of next_blocks should lead to head
    next_blocks: List["Block"] = field(default_factory=list)
    _hash: Optional[str] = None

    def hash(self, recompute=False):
        if not self._hash or recompute:
            if self.height == 0:
                self._hash = ""
            else:
                hasher = hashlib.sha256()
                hasher.update(str(self.height).encode())
                hasher.update(self.prev_hash.encode())
                for stx in self.signed_transactions:
                    hasher.update(stx.signature.encode())
                hasher.update(str(self.nonce).encode())
                self._hash = hasher.hexdigest()
        return self._hash

    def __str__(self):
        return f"""BLOCK(height: {self.height}, prev: {self.prev_hash},
        nonce: {self.nonce}, SignedTransactions{self.signed_transactions})"""

    def html(self):
        hash = self.hash()[:20]
        str_stx = ""
        for tx in self.signed_transactions:
            str_stx += f"<li>{tx.html()}</li>"

        return f"""<div style='border:1px solid black; padding:5px; margin:5px;'>
        <b>height</b>: {self.height}, <b>hash</b>: {hash}, <b>nonce</b>: {self.nonce},<br>
        <b>signed transactions</b>: <ul style='margin:0px'>{str_stx}</ul>
        </div>"""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['babbagecoin',
 'babbagecoin.client',
 'babbagecoin.common',
 'babbagecoin.master',
 'babbagecoin.miner']

package_data = \
{'': ['*']}

install_requires = \
['Flask>=2.0.3,<3.0.0',
 'cryptography>=36.0.1,<37.0.0',
 'marshmallow>=3.14.1,<4.0.0',
 'python-dotenv>=0.19.2,<0.20.0',
 'requests>=2.27.1,<3.0.0',
 'sentry-sdk[flask]>=1.5.8,<2.0.0']

setup_kwargs = {
    'name': 'babbagecoin',
    'version': '0.4.0',
    'description': 'Second edition of Project Babbage, create a blockchain from scratch.',
    'long_description': "# Babbagecoin\n\n### A understandable proof of work blockchain.\n\n**Visit our [web page](https://projectbabbage.github.io/babbagecoin/) for a nice general overview of the project !**\n\n## Quickly run a node !\n\n\n```bash\npip install --user babbagecoin\n```\n\n```bash\npython -m babbagecoin master\n# then in another terminal:\npython -m babbagecoin miner\n```\n\n# For development\n\n## Requirements\n\nInstall `docker`, `docker-compose` and `python3.8` or above.\nYou also need `poetry`.\n\nRun `poetry install` then `poetry shell`\n\n## Launch\n\nRun the node (master + miner):\n\n`make`\n\nStop the node properly:\n\n`make stop` --> to stop all containers if you didn't stopped them properly (if you did two Ctrl+C in a row)\n\nOr you can run separately the master or miner:\n\n`make (master | miner)`\n\nThere are also VSCode actions for debugging each one of them.\n\n## Interact with the blockchain\n\n!! Important\n\nConfigure your blockchain by creating a `.env` file on the same model as what's in `.env.example`.\n\nThe wallet will generate a private key for you, save it to private.key.<CURRENT_USER>, and reuse it afterward. A public key is derived from this private key. The hash of the public key is your address, for example e93417c7 (first 8 characters).\nThe wallet is managing only one private key at the time.\n\n### Transactions\n\nUse the `bbc.sh` script (`chmod +x` it first):\n\n`./bbc.sh tx MARTIAL 10 0.3` --> sending 10BBC with 0.5BBC fees to MARTIAL\n\n_`make tx` is a shortcut for the above command_\n\n### Balance\n\n`./bbc.sh balance` to get your wallet balance\n\n_`make balance` is a shortcut for the above command_\n",
    'author': 'Quentin Garchery',
    'author_email': 'garchery.quentin@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ProjectBabbage/babbagecoin',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from pathlib import Path

from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.exceptions import InvalidSignature

from babbagecoin.common.models import MINING_REWARD_ADDRESS, PubKey, Transaction, SignedTransaction
from babbagecoin.common.exceptions import InvalidSignatureForTransaction
from babbagecoin.common.context import get_current_user


class Wallet:
    private_key: RSAPrivateKey

    def __init__(self, load_from_file=True, save_to_file=True):
        self.private_key = rsa.generate_private_key(public_exponent=65537, key_size=512)

        user_privk_filepath = f"private.key.{get_current_user()}"
        if load_from_file and Path(user_privk_filepath).is_file():
            self.private_key = Wallet.load_priv_keys(user_privk_filepath)
            print(f"Found {user_privk_filepath}, using it.")
        if save_to_file and not Path(user_privk_filepath).is_file():
            self.save_files()
            print("Saving key files.")

    def save_files(self):
        """Create the private.key.<CURRENT_USER> and public.key.<CURRENT_USER> files."""
        user_private = f"private.key.{get_current_user()}"
        user_public = f"public.key.{get_current_user()}"
        with open(user_private, "wb") as priv:
            priv.write(self.decode_private_key())
        with open(user_public, "wb") as pub:
            pub.write(self.decode_public_key())

    def get_public_key(self) -> PubKey:
        return PubKey(self.private_key.public_key())

    def decode_public_key(self) -> bytes:
        return self.private_key.public_key().public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )

    def decode_private_key(self) -> bytes:
        return self.private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )

    def sign(self, transaction: Transaction) -> str:
        return self.private_key.sign(
            transaction.hash().encode(),
            padding.PSS(mgf=padding.MGF1(hashes.SHA256()), salt_length=padding.PSS.MAX_LENGTH),
            hashes.SHA256(),
        ).hex()

    @staticmethod
    def verify_signature(signed_transaction: SignedTransaction) -> bool:

        signature = bytes.fromhex(signed_transaction.signature)
        transaction = signed_transaction.transaction
        transaction_hash = transaction.hash().encode()

        if transaction.sender.dumps() != MINING_REWARD_ADDRESS:
            try:
                signed_transaction.transaction.sender.rsa_pub_key.verify(
                    signature,
                    transaction_hash,
                    padding.PSS(
                        mgf=padding.MGF1(hashes.SHA256()),
                        salt_length=padding.PSS.MAX_LENGTH,
                    ),
                    hashes.SHA256(),
                )
            except InvalidSignature:
                raise InvalidSignatureForTransaction(
                    f"INVALID SIGNATURE {signed_transaction.signature}, for stx "
                    "{signed_transaction}, of tx hash: {transaction_hash}"
                )

    @staticmethod
    def load_pub_key(filepath):
        with open(filepath, "rb") as bf:
            return PubKey.load_from_bytes(bf.read())

    @staticmethod
    def load_from_file(filepath):
        w = Wallet(load_from_file=False, save_to_file=False)
        w.private_key = Wallet.load_priv_keys(filepath)
        return w

    @staticmethod
    def load_priv_keys(filepath):
        with open(filepath, "rb") as fkey:
            return load_pem_private_key(fkey.read(), password=None)


wallet = Wallet()



from babbagecoin.common.exceptions import (
    InvalidBlockHash,
    InvalidBlockHeight,
    InvalidSignatureForTransaction,
    RewardTransactionNotUnique,
    BadRewardTransaction,
)
from babbagecoin.common.models import Block, MINING_REWARD_AMOUNT, MINING_REWARD_ADDRESS
from babbagecoin.common.block_service import is_block_hash_valid
from babbagecoin.common.wallet import Wallet
from babbagecoin.master.transaction_service import (
    refresh_transactions_from_old_block,
    refresh_transactions_from_new_block,
    get_reward_transaction,
    validated_transactions,
)

block_tbl = {}  # map from hash to block
genesis: Block = Block(prev_hash=None, height=0)
block_tbl[genesis.hash()] = genesis
head: Block = genesis


def update_block_tbl_between(start, end):
    # requires end path invariant
    block_tbl[start.hash()] = start
    if start != end:
        update_block_tbl_between(start.next_blocks[0], end)


def remove_block_tbl_between(start, end):
    block_tbl.pop(end.hash())
    if start != end:
        prev = block_tbl[end.prev_hash]
        remove_block_tbl_between(start, prev)


def find_common_ancestor_of(leaf1, leaf2):
    if leaf1.height < leaf2.height:
        temp = leaf2
        leaf2 = leaf1
        leaf1 = temp
    cursor1 = leaf1
    while cursor1.height > leaf2.height + 1:
        cursor1 = block_tbl[cursor1.prev_hash]
    cursor2 = leaf2
    while cursor1.prev_hash != cursor2.hash():
        cursor1 = block_tbl[cursor1.prev_hash]
        cursor2 = block_tbl[cursor2.prev_hash]
    return cursor2


def make_primary_between(block, next_block):
    """
      block
      ┌───┐   ┌───┐                       ┌───┐   ┌───┐
    ──┤ s ├───┤ y │                     ──┤ s ├─┬─┤ e │
      └───┘   └───┘                       └───┘ │ └───┘
               next_block   ─────────►          │
              ┌───┐                             │ ┌───┐
             ─┤ e │                             └─┤ y │
              └───┘                               └───┘
    """

    def no_next(b: Block):
        return b.hash() != next_block.hash()

    block.next_blocks = [next_block] + list(filter(no_next, block.next_blocks))


def sane_from(start: Block, starting_height: int):
    """
    Run verify_block on each block from start.
    Returns False if an inconsistency is detected:
    - InvalidBlockHash: the block hash doesn't match the required difficulty
    - InvalidBlockHeight: the block height does not get incremented by 1 every time
    - BadRewardTransaction: the reward transaction as bad amount or bad address
    - RewardTransactionNotUnique: the reward transaction address appears after the first tx
    - InvalidSignature: the signature for a stx is incorrect
    """

    def verify_block(block: Block, expected_height: int):
        if block.height != expected_height:
            raise InvalidBlockHeight(
                f"Expected height {expected_height}, actual height {block.height}"
            )
        if not is_block_hash_valid(block):
            raise InvalidBlockHash(f"INVALID BLOCK HASH {block.hash()}")
        for i, stx in enumerate(block.signed_transactions):
            if i == 0 and (
                stx.transaction.sender.dumps() != MINING_REWARD_ADDRESS
                or stx.transaction.amount != MINING_REWARD_AMOUNT
            ):
                raise BadRewardTransaction(f"BAD REWARD TRANSACTION {stx}")
            if i != 0 and stx.transaction.sender.dumps() == MINING_REWARD_ADDRESS:
                raise RewardTransactionNotUnique(f"REWARD TRANSACTION NOT UNIQUE {stx}")

            Wallet.verify_signature(stx)  # can raise an InvalidSignatureForTransaction

    try:
        b = start
        expected_height = starting_height
        verify_block(b, expected_height)
        while len(b.next_blocks) != 0:
            b = b.next_blocks[0]
            expected_height += 1
            verify_block(b, expected_height)
        return True
    except (
        InvalidBlockHash,
        InvalidBlockHeight,
        BadRewardTransaction,
        RewardTransactionNotUnique,
        InvalidSignatureForTransaction,
    ) as e:
        print(e)
        return False


def refresh_transactions_switch(start: Block, ancestor: Block, end: Block):
    # take into account transactions of the old branch start
    b = start
    while b.hash() != ancestor.hash():
        refresh_transactions_from_old_block(b)
        b = block_tbl[b.prev_hash]
    # take into account transactions from the new branch end
    b = end
    while b.hash() != ancestor.hash():
        refresh_transactions_from_new_block(b)
        b = block_tbl[b.prev_hash]


def update_blockchain(anchor: Block, leaf: Block):
    """
    ancestor                      head
     ┌───┐   ┌───┐               ┌───┐
     │ a ├─┬─┤ b ├───────────────┤ c │
     └───┘ │ └───┘               └───┘
           │
           │ ┌───┐   ┌───┐
           └─┤ d ├─┬─┤ e │
             └───┘ │ └───┘
                   │         anchor        leaf
                   │ ┌───┐   ┌───┐         ┌───┐
                   └─┤ f │   │ g ├─────────┤ h │
                     └───┘   └───┘         └───┘
                  │
                  │
                  ▼
    ancestor                 anchor      head = leaf
     ┌───┐   ┌───┐   ┌───┐   ┌───┐         ┌───┐
     │ a ├─┬─┤ d ├─┬─┤ f ├───┤ g ├─────────┤ h │
     └───┘ │ └───┘ │ └───┘   └───┘         └───┘
           │       │
           │       │ ┌───┐
           │       └─┤ e │
           │         └───┘
           │
           │ ┌───┐               ┌───┐
           └─┤ b ├───────────────┤ c │
             └───┘               └───┘
    """
    global head
    if anchor.prev_hash not in block_tbl:
        raise Exception("Does not connect to our blockchain")
    anchor_prev = block_tbl[anchor.prev_hash]
    if leaf.height > head.height and sane_from(anchor, anchor_prev.height + 1):
        update_block_tbl_between(anchor, leaf)
        ancestor = find_common_ancestor_of(leaf, head)
        refresh_transactions_switch(head, ancestor, anchor_prev)
        b = anchor
        while True:
            excess_transactions = refresh_transactions_from_new_block(b)
            if b.hash() == leaf.hash() or excess_transactions != set():
                break
            b = b.next_blocks[0]
        if excess_transactions == set():
            # enforce the head path invariant
            make_primary_between(anchor_prev, anchor)
            # change head to be the new leaf
            head = leaf
            print(f"Changing head to {head.hash()}")
        else:
            refresh_transactions_switch(b, ancestor, head)
            for stx in excess_transactions:
                validated_transactions.add(stx)
            remove_block_tbl_between(anchor, leaf)
            print("Discarding new blocks due to a transaction already validated")
    """
    Verification of a transaction already validated (b) ending up in rejecting the incoming blocks:

    ancestor            anchor    leaf                      │
     ┌───┐    ┌───┐     ┌───┐     ┌───┐          ┌───┐    ┌─▼─┐     ┌───┐     ┌───┐
     │ b │    │ f │     │ a │     │ c │          │ b │    │ f │     │ a │     │ c │
     │ e ├────┤   │     │ b ├─────┤ d │          │ e ├────┤   │     │ b ├─────┤ d │
     └──┬┘    └───┘     └───┘     └───┘          └──┬┘    └───┘     └───┘     └───┘
        │              head                         │
        │     ┌───┐   ┌───┐                         │     ┌───┐   ┌───┐
        └─────┤ r ├───┤ x │                         └─────┤ r ├───┤ x │
              │ t │   │ y │                               │ t │   │ y │
              └───┘   └─▲─┘            ────────►          └───┘   └───┘
                        │

         validated     mempool                       validated     mempool
         ┌─────┐       ┌─────┐                       ┌─────┐       ┌─────┐
         │     │       │     │                       │     │       │     │
         │  b  │       │  a  │                       │  b  │       │  a  │
         │  e  │       │  c  │                       │  e  │       │  c  │
         │  r  │       │  f  │                       │  f  │       │     │
         │  t  │       │     │                       │     │       │  r  │
         │  x  │       │     │                       │     │       │  t  │
         │  y  │       │     │                       │     │       │  x  │
         │     │       │     │                       │     │       │  y  │
         └─────┘       └─────┘                       └─────┘       └─────┘
                                                                │
                                                                │
                                                                ▼
                                                                      │
     ┌───┐    ┌───┐                              ┌───┐    ┌───┐     ┌─▼─┐     ┌───┐
     │ b │    │ f │                              │ b │    │ f │     │ a │     │ c │
     │ e ├────┤   │                              │ e ├────┤   │     │ b ├─────┤ d │
     └──┬┘    └───┘                              └──┬┘    └───┘     └───┘     └───┘
        │                                           │
        │     ┌───┐   ┌───┐                         │     ┌───┐   ┌───┐
        └─────┤ r ├───┤ x │                         └─────┤ r ├───┤ x │
              │ t │   │ y │            ◄────────          │ t │   │ y │
              └───┘   └─▲─┘                               └───┘   └───┘
                        │

         validated     mempool                       validated     mempool     excess
         ┌─────┐       ┌─────┐                       ┌─────┐       ┌─────┐     ┌─────┐
         │     │       │     │                       │     │       │     │     │     │
         │  b  │       │  a  │                       │  b  │       │     │     │  b  │
         │  e  │       │  c  │                       │  e  │       │  c  │     │     │
         │  r  │       │  f  │                       │  f  │       │     │     │     │
         │  t  │       │     │                       │  a  │       │  r  │     │     │
         │  x  │       │     │                       │     │       │  t  │     │     │
         │  y  │       │     │                       │     │       │  x  │     │     │
         │     │       │     │                       │     │       │  y  │     │     │
         └─────┘       └─────┘                       └─────┘       └─────┘     └─────┘

    """


def get_head() -> Block:
    return head


def build_working_block() -> Block:
    rewardTransaction = get_reward_transaction()
    new_block = Block(
        height=head.height + 1,
        prev_hash=head.hash(),
        signed_transactions=[rewardTransaction],
    )
    return new_block



import json

import requests

from babbagecoin.common.models import Block, SignedTransaction
from babbagecoin.common.schemas import BlockSchema, SignedTransactionSchema
from babbagecoin.common.context import NetworkContext

context = NetworkContext()


def broadcast_block(block: Block):
    json_block_dict = {"url": context.myUrl, "block": BlockSchema.dump(block)}
    print(f"Broacasting block {block.hash()}")
    for host in context.known_nodes:
        if host != context.myIp:
            try:
                requests.post(
                    f"http://{host}:5000/blocks/updateblock",
                    json.dumps(json_block_dict),
                    headers={"Content-Type": "application/json"},
                )
            except requests.exceptions.ConnectionError:
                print(f"Cannot connect to node {host}")
            except Exception as e:
                print(f"An error occured when broadcasting the block: {e}")


def broadcast_transaction(signed_transaction: SignedTransaction):
    signed_transaction_json = SignedTransactionSchema.dumps(signed_transaction)
    print(f"Broadcasting transaction {signed_transaction.hash()}")
    for host in context.known_nodes:
        if host != context.myIp:
            try:
                requests.post(
                    f"http://{host}:5000/transactions/add",
                    signed_transaction_json,
                    headers={"Content-Type": "application/json"},
                )
            except requests.exceptions.ConnectionError:
                print(f"Cannot connect to node {host}")
            except Exception as e:
                print(f"Cannot broadcast transaction: {e}")
